from flask import Flask, render_template, request, redirect
from shlex import split
from datetime import datetime
from collections import Counter
from matplotlib.pyplot import cm
import pandas as pd
import json

app = Flask(__name__)
th = None

names = {
    "path_length": "Path length",
    "directory": "Directory",
    "exec": "Exec name",
    "hash": "Exec hash",
    "producer": "Editor name",
    "time": "Timestamp",
    "pid": "PID",
    "guid": "GUID",
    "command_line": "Command line",
}

titles = {
    "path_length": "Directories of size <b>{val}</b>",
    "directory": "Programs executed in <b>{val}</b>",
    "exec": "Results for <b>{val}</b>",
    "process_exec": "Results for <b>{val}</b>",
    "hash": "Programs with hash <b>{val}</b>",
    "producer": "Programs edited by <b>{val}</b>",
    "time": "Processes executed at <b>{val}</b>",
    "pid": "Processes with PID <b>{val}</b>",
    "guid": "Processes with GUID <b>{val}</b>",
    "command_line": "Command lines containing <b>{val}</b>"
}

tooltips = {
    "hf": "Hash Frequency",
    "df": "Directory Frequency",
    "pl": "Path Length (depth)",
    "nf": "Name Frequency",
    "nc": "Number of Children processes",
    "ct": "Company Trust level",
    "nd": "Number of distinct directories for this program"
}

cmap = cm.Spectral
colormap_single = lambda x: cmap(1 - x/10., bytes=True)
colormap = lambda x: list(map(tuple, cmap(1 - x/10., bytes=True)))

def run_app(threat_hunter, **kwargs):
    global th
    th = threat_hunter
    app.run(**kwargs)




@app.route("/")
def index():
    scores, _ = th.get_results(limit=25)
    return render_template("index.html",
                           scores=scores[:25])



@app.route("/about")
def about():
    return render_template("about.html")


@app.route("/process", methods=["GET"])
def process():
    id = request.args.get("id")
    if id is None : return redirect("/")
    node, children, parents, hashes, producers, directories, links = th.con.retrieve_process(int(id))
    print(node)
    labels = set(node.labels) - set(["process"])
    computer = list(labels)[0]
    date = datetime.strptime(node["time"], "%Y-%m-%d %H:%M:%S.%f CET")
    date_str = date.strftime("%a. %d %B %Y - %H:%M %Ss")

    graph = {"nodes": [{"id": node.id, "type": "this process", "title": node["process_exec"]}] \
                    + [{"id": n.id, "type": "child", "title": n["process_exec"]} for n in children] \
                    + [{"id": n.id, "type": "parent", "title": n["process_exec"]} for n in parents],
             "links": links
            }
    return render_template("process.html",
                           node=node,
                           date=date_str,
                           host=computer,
                           graph=graph,
                           producer=producers[0]["name"],
                           hash=hashes[0]["hash"],
                           directory=directories[0]["name"],
                           cmap=colormap_single)


@app.route("/postTicket", methods=["POST"])
def post_ticket():
    id = request.form.get("id")
    if id is None :
        return json.dumps({"status": "Missing ID parameter"}), 400
    th.con.mark_as_a_threat(id)
    return json.dumps({"status": "OK"}), 200
    


@app.route("/stats", methods=["GET"])
def stats():
    criterion = request.args.get("criterion")
    value = request.args.get("val")
    if criterion == "threat" :
        return threat_score(request)
    elif criterion in names :
        if value is not None :
            return valued_attribute_counter(request, criterion, value)
        return attribute_counter(request, th.criteria[criterion].items(), names[criterion])
    else :
        return redirect("/stats?criterion=threat")
    return render_template("stats.html")



def manage_page(results, p):
    p_max = 1 + (len(results)-1)//50
    if p < 1 : p = p_max
    elif p > p_max : p = p_max
    return p


def threat_score(request):
    p = int(request.args.get("p", "1"))
    results, _ = th.get_results(limit=p*50)
    p = manage_page(results, p)
    args = dict(request.args)
    args["p"] = p

    df = pd.DataFrame(results[50*(p-1):50*p])
    df["_link"] = "/process?id=" + df["_id"].astype("string")
    df["_color"] = "rgb" + pd.Series(colormap(df["threat"])).astype("string")
    df = df[["threat", "process_exec", "process_path", "producer", "number_child", "_color", "_link"]]
    df.columns = ["Score", "Exec name", "Process path", "Editor", "Children", "_color", "_link"]
    cnt = Counter(list(map(lambda t: round(t["threat"], 1), results)))
    keys = list(cnt.keys())
    values = list(cnt.values())
    arange = list(range(len(values)))
    return render_template("stats.html",
                           results=df,
                           keys=keys,
                           values=values,
                           arange=arange,
                           args=args)


def attribute_counter(request, items, name):
    results = sorted(items, key=lambda t:t[1], reverse=True)
    p = int(request.args.get("p", "1"))
    p = manage_page(results, p)
    args = dict(request.args)
    args["p"] = p

    df = pd.DataFrame([{name: t[0], "Count": t[1]} for t in results[50*(p-1):50*p]])
    df["_link"] = "/stats?criterion=" + args["criterion"] + "&val=" + df[name].astype("string").map(lambda s: s.replace("\\", "\\\\"))
    keys = list(map(lambda t:t[0], results))
    values = list(map(lambda t:t[1], results))
    arange = list(range(len(values)))
    return render_template("stats.html",
                           results=df,
                           keys=keys,
                           values=values,
                           arange=arange,
                           args=args)


def valued_attribute_counter(request, criterion, value) :
    other_criteria = ("time", "pid", "guid", "command_line")
    if criterion == "exec":
        if value.split(":")[0] in other_criteria :
            by = value.split(":")[0]
            val = ":".join(value.split(":")[1:])
            return redirect(f"/stats?criterion={by}&val={val}")
        return search(request, value)
    if criterion in other_criteria:
        return search(request, value, by=criterion)
    results = th.con.retrieve_process_by_attribut(criterion, value)
    p = int(request.args.get("p", "1"))
    p = manage_page(results, p)
    args = dict(request.args)
    args["p"] = p
    title = titles[criterion].format(val=value)

    df = pd.DataFrame(results[50*(p-1):50*p])
    if criterion == "path_length" :
        df["_link"] = "/stats?criterion=directory&val=" + df["process_exec"].astype("string").map(lambda s: s.replace("\\", "\\\\"))
        df.columns = ["Directory", "Count", "_link"]
    else :
        df["_link"] = "/stats?criterion=exec&val=" + df["process_exec"].astype("string")
        df.columns = ["Exec name", "Count", "_link"]
    keys = list(map(lambda t:t["process_exec"], results))
    values = list(map(lambda t:t["count"], results))
    arange = list(range(len(values)))
    return render_template("stats.html",
                           results=df,
                           keys=keys,
                           values=values,
                           arange=arange,
                           args=args,
                           title=title)


def search(request, query, by="process_exec") :
    results = th.con.search_processes(query, by=by)
    results = sorted(results, key=lambda t:t["threat"], reverse=True)
    p = int(request.args.get("p", "1"))
    p = manage_page(results, p)
    args = dict(request.args)
    args["p"] = p
    title = titles[by].format(val=query)

    df = pd.DataFrame(results[50*(p-1):50*p])
    if len(results) > 0 :
        df["_link"] = "/process?id=" + df["_id"].astype("string")
        df["_color"] = "rgb" + pd.Series(colormap(df["threat"])).astype("string")
        df = df[["threat", "process_exec", "process_path", "pid", "producer", "_color", "_link"]]
        df.columns = ["Score", "Exec name", "Process path", "PID", "Editor", "_color", "_link"]
    cnt = Counter(list(map(lambda t: round(t["threat"], 1), results)))
    keys = list(cnt.keys())
    values = list(cnt.values())
    arange = list(range(len(values)))
    return render_template("stats.html",
                           results=df,
                           keys=keys,
                           values=values,
                           arange=arange,
                           args=args,
                           title=title)



@app.route("/ticket-basket", methods=["GET"])
def get_tickets():
    results = th.con.get_threats()
    results = sorted(results, key=lambda t:t["heuristic"], reverse=True)
    p = int(request.args.get("p", "1"))
    p = manage_page(results, p)
    args = dict(request.args)
    args["p"] = p

    df = pd.DataFrame(results[50*(p-1):50*p])
    if len(results) > 0 :
        df["_link"] = "/process?id=" + df["_id"].astype("string")
        df["_color"] = "rgb" + pd.Series(colormap(df["heuristic"])).astype("string")
        df = df[["heuristic", "process_exec", "process_path", "pid", "producer", "hf", "df", "pl", "nf", "nc", "ct", "nd", "_color", "_link"]]
        df.columns = ["Score", "Exec name", "Process path", "PID", "Editor", "hf", "df", "pl", "nf", "nc", "ct", "nd", "_color", "_link"]
    cnt = Counter(list(map(lambda t: round(t["heuristic"], 1), results)))
    keys = list(cnt.keys())
    values = list(cnt.values())
    return render_template("threats.html",
                           req="basket",
                           base_page="/ticket-basket?",
                           results=df,
                           tooltips=tooltips,
                           keys=keys,
                           values=values,
                           args=args)



@app.route("/clusters", methods=["GET"])
def get_clusters():
    c = int(request.args.get("c", "0"))
    print("Chargement des clusters...")
    clusters = th.clusters
    if clusters is None :
        _, clusters = th.get_results()
    c %= len(clusters)
    mean_score, results = clusters[c]
    print("Clusters chargés !")

    p = int(request.args.get("p", "1"))
    p = manage_page(results, p)
    args = dict(request.args)
    args["p"] = p

    df = pd.DataFrame(results[50*(p-1):50*p])
    if len(results) > 0 :
        df["_link"] = "/process?id=" + df["_id"].astype("string")
        df["_color"] = "rgb" + pd.Series(colormap(df["heuristic"])).astype("string")
        df = df[["heuristic", "process_exec", "process_path", "pid", "producer", "hf", "df", "pl", "nf", "nc", "ct", "nd", "_color", "_link"]]
        df.columns = ["Score", "Exec name", "Process path", "PID", "Editor", "hf", "df", "pl", "nf", "nc", "ct", "nd", "_color", "_link"]
    cnt = Counter(list(map(lambda t: round(t["heuristic"], 1), results)))
    keys = list(cnt.keys())
    values = list(cnt.values())
    return render_template("threats.html",
                           req="clusters",
                           base_page=f"/clusters?c={c}&",
                           c=c,
                           n=len(results),
                           results=df,
                           tooltips=tooltips,
                           keys=keys,
                           values=values,
                           args=args)
